.md-content {
    margin-right: 2em;
}
.md-main__inner {
    padding: 0;
}
.navheader {
    padding-left: 1em;
}
.md-sidebar--secondary {
    display: none;
}
.viewcode-link {
    font-size: .75em;
    margin-left: .5em;
}
.caption-text {
    font-style: oblique;
}
code.descname {
    background: transparent;
    box-shadow: none;
    font-size: 100% !important;
}
code.descname {
    background: transparent;
    box-shadow: none;
    font-size: 100% !important;
    font-weight: bold;
}
code.descclassname {
    background: transparent;
    box-shadow: none;
    font-size: 100% !important;
    font-weight: bold;
}
.rst-versions {
    font-size: 1.75em;
}
